
import shutil 
import json 
import zipfile 
import io 
import copy
import warnings
from pathlib import Path 

import torch 
import numpy as np 

import pytorch_lightning as pl 
from scipy.sparse import csc_matrix 
from pytorch_tabnet.utils import (
    create_explain_matrix,
    ComplexEncoder,
)
import torch.nn.functional as F
from pytorch_tabnet.tab_network import TabNet
from torchmetrics.functional.classification.stat_scores import _stat_scores_update
from tqdm import tqdm 
from typing import *
from metrics import aggregate_metrics

class SIMSClassifier(pl.LightningModule):
    def __init__(
        self,
        input_dim,
        output_dim,
        n_d=8,
        n_a=8,
        n_steps=3,
        gamma=1.3,
        cat_idxs=[],
        cat_dims=[],
        cat_emb_dim=1,
        n_independent=2,
        n_shared=2,
        epsilon=1e-15,
        virtual_batch_size=128,
        momentum=0.02,
        mask_type="sparsemax",
        lambda_sparse = 1e-3,
        optim_params: Dict[str, float]=None,
        metrics: Dict[str, Callable]=None,
        scheduler_params: Dict[str, float]=None,
        weights: torch.Tensor=None,
        loss: Callable=None, # will default to cross_entropy
        pretrained: bool=None,
        no_explain: bool=False,
    ) -> None:
        super().__init__()

        # Stuff needed for training
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.lambda_sparse = lambda_sparse

        self.optim_params = optim_params
        
        self.weights = weights 
        self.loss = loss 

        if pretrained is not None:
            self._from_pretrained(**pretrained.get_params())

        if metrics is None:
            self.metrics = aggregate_metrics(num_classes=self.output_dim)
        else:
            self.metrics = metrics 

        if optim_params is None:
            self.optim_params = {
                'optimizer': torch.optim.Adam,
                'lr': 0.001,
                'weight_decay': 0.001,
            }
        else:
            self.optim_params = optim_params

        if scheduler_params is None:
            self.scheduler_params={
                'scheduler': torch.optim.lr_scheduler.ReduceLROnPlateau,
                'factor': 0.75, # Reduce LR by 25% on plateau
            }
        else:
            self.scheduler_params = scheduler_params

        print(f'Initializing network')
        self.network = TabNet(
            input_dim=input_dim, 
            output_dim=output_dim, 
            n_d=n_d,
            n_a=n_a,
            n_steps=n_steps,
            gamma=gamma,
            cat_idxs=cat_idxs,
            cat_dims=cat_dims,
            cat_emb_dim=cat_emb_dim,
            n_independent=n_independent,
            n_shared=n_shared,
            epsilon=epsilon,
            virtual_batch_size=virtual_batch_size,
            momentum=momentum,
            mask_type=mask_type,
        )

        print(f'Initializing explain matrix')
        if not no_explain:
            self.reducing_matrix = create_explain_matrix(
                self.network.input_dim,
                self.network.cat_emb_dim,
                self.network.cat_idxs,
                self.network.post_embed_dim,
            )

    def forward(self, x):
        return self.network(x)

    def _compute_loss(self, y, y_hat):
        # If user doesn't specify, just set to cross_entropy
        if self.loss is None:
            self.loss = F.cross_entropy 

        return self.loss(y, y_hat, weight=self.weights)

    def _compute_metrics(self, 
        y_hat: torch.Tensor, 
        y: torch.Tensor, 
        tag: str,
        on_epoch=True, 
        on_step=False,
    ):
        for name, metric in self.metrics.items():
            val = metric(y_hat, y)
            self.log(
                f"{tag}_{name}", 
                val, 
                on_epoch=on_epoch, 
                on_step=on_step,
                logger=True,
            )

    def _step(self, batch):
        x, y = batch
        y_hat, M_loss = self.network(x)

        loss = self._compute_loss(y_hat, y)

        # Add the overall sparsity loss
        loss = loss - self.lambda_sparse * M_loss
        return y, y_hat, loss


    def _step(self, batch, tag):
        x, y = batch
        y_hat, M_loss = self.network(x)

        loss = self._compute_loss(y_hat, y)
        # Add the overall sparsity loss
        loss = loss - self.lambda_sparse * M_loss

        self.log(f"{tag}_loss", loss, logger=True, on_epoch=True, on_step=True)
        self._compute_metrics(y_hat, y, tag)
        
        tp, fp, _, fn = _stat_scores_update(
            preds=y_hat,
            target=y,
            num_classes=self.output_dim,
            reduce="macro",
        )

        return {
            "loss": loss,
            "tp": tp,
            "fp": fp,
            "fn": fn,
        }

    # Calculations on step
    def training_step(self, batch, batch_idx):
        return self._step(batch, 'train')

    def validation_step(self, batch, batch_idx):
        return self._step(batch, 'val')

    def test_step(self, batch, batch_idx):
        return self._step(batch, 'test')
    
    def _epoch_end(self, step_outputs, tag):
        tps, fps, fns = [], [], []
        
        for i in range(len(step_outputs)):
            res = step_outputs[i]
            tp, fp, fn = res['tp'], res['fp'], res['fn']
                
            tps.append(tp.cpu().numpy())
            fps.append(fp.cpu().numpy())
            fns.append(fn.cpu().numpy())
            
        tp = np.sum(np.array(tps), axis=0)
        fp = np.sum(np.array(fps), axis=0)
        fn = np.sum(np.array(fns), axis=0)
        
        precision = tp / (tp + fp)
        recall = tp / (tp + fn)
        f1s = 2*(precision * recall) / (precision + recall)
        f1s = np.nan_to_num(f1s)

        self.log(
            f"{tag}_median_f1", 
            np.nanmedian(f1s), 
            logger=True, 
            on_step=False, 
            on_epoch=True
        )

        return f1s 

    # Calculation on epoch end, for "median F1 score"
    def training_epoch_end(self, step_outputs):
        self._epoch_end(step_outputs,'train')
        
    def validation_epoch_end(self, step_outputs):
        self._epoch_end(step_outputs, 'val') 
    
    def test_epoch_end(self, step_outputs):
        self._epoch_end(step_outputs, 'test') 

    def configure_optimizers(self):
        if 'optimizer' in self.optim_params:
            optimizer = self.optim_params.pop('optimizer')
            optimizer = optimizer(self.parameters(), **self.optim_params)
        else:
            optimizer = torch.optim.Adam(self.parameters(), lr=0.2, weight_decay=1e-5)

        if self.scheduler_params is not None:
            scheduler = self.scheduler_params.pop('scheduler')
            scheduler = scheduler(optimizer, **self.scheduler_params)

        if self.scheduler_params is None:
            return optimizer
        
        return {
            'optimizer': optimizer,
            'lr_scheduler': scheduler,
            'monitor': 'train_loss',
        }

    def explain(self, loader, cache=False, normalize=False):
        if cache and self._explain_matrix is not None:
            return self._explain_matrix 
            
        self.network.eval()
        res_explain = []

        for batch_nb, data in enumerate(tqdm(loader)):
            if isinstance(data, tuple): # if we are running this on already labeled pairs and not just for inference
                data, _ = data 
                
            M_explain, masks = self.network.forward_masks(data)
            for key, value in masks.items():
                masks[key] = csc_matrix.dot(
                    value.cpu().detach().numpy(), self.reducing_matrix
                )

            original_feat_explain = csc_matrix.dot(
                M_explain.cpu().detach().numpy(),
                self.reducing_matrix
            )
            res_explain.append(original_feat_explain)

            if batch_nb == 0:
                res_masks = masks
            else:
                for key, value in masks.items():
                    res_masks[key] = np.vstack([res_masks[key], value])

        res_explain = np.vstack(res_explain)
        
        if normalize:
            res_explain /= np.sum(res_explain, axis=1)[:, None]

        if cache:
            self._explain_matrix = res_explain

        return res_explain, res_masks

    def _compute_feature_importances(self, dataloader):
        M_explain, _ = self.explain(dataloader, normalize=False)
        sum_explain = M_explain.sum(axis=0)
        feature_importances_ = sum_explain / np.sum(sum_explain)
        
        return feature_importances_

    def feature_importances(self, dataloader, cache=False):
        if cache and self._feature_importances is not None:
            return self._feature_importances
        else:
            f = self._compute_feature_importances(dataloader)
            if cache:
                self._feature_importances = f 
            return f 
    
    def save_model(self, path):
        saved_params = {}
        init_params = {}
        for key, val in self.get_params().items():
            if isinstance(val, type):
                # Don't save torch specific params
                continue
            else:
                init_params[key] = val
        saved_params["init_params"] = init_params

        class_attrs = {
            "preds_mapper": self.preds_mapper
        }
        saved_params["class_attrs"] = class_attrs

        # Create folder
        Path(path).mkdir(parents=True, exist_ok=True)

        # Save models params
        with open(Path(path).joinpath("model_params.json"), "w", encoding="utf8") as f:
            json.dump(saved_params, f, cls=ComplexEncoder)

        # Save state_dict
        torch.save(self.network.state_dict(), Path(path).joinpath("network.pt"))
        shutil.make_archive(path, "zip", path)
        shutil.rmtree(path)
        print(f"Successfully saved model at {path}.zip")
        return f"{path}.zip"

    def load_model(self, filepath):
        try:
            with zipfile.ZipFile(filepath) as z:
                with z.open("model_params.json") as f:
                    loaded_params = json.load(f)
                    loaded_params["init_params"]["device_name"] = self.device_name
                with z.open("network.pt") as f:
                    try:
                        saved_state_dict = torch.load(f, map_location=self.device)
                    except io.UnsupportedOperation:
                        # In Python <3.7, the returned file object is not seekable (which at least
                        # some versions of PyTorch require) - so we'll try buffering it in to a
                        # BytesIO instead:
                        saved_state_dict = torch.load(
                            io.BytesIO(f.read()),
                            map_location=self.device,
                        )
        except KeyError:
            raise KeyError("Your zip file is missing at least one component")

        self.__init__(**loaded_params["init_params"])

        self._set_network()
        self.network.load_state_dict(saved_state_dict)
        self.network.eval()
        self.load_class_attrs(loaded_params["class_attrs"])

    def load_weights_from_unsupervised(self, unsupervised_model):
        update_state_dict = copy.deepcopy(self.network.state_dict())
        for param, weights in unsupervised_model.network.state_dict().items():
            if param.startswith("encoder"):
                # Convert encoder's layers name to match
                new_param = "tabnet." + param
            else:
                new_param = param
            if self.network.state_dict().get(new_param) is not None:
                # update only common layers
                update_state_dict[new_param] = weights

    def _from_pretrained(self, **kwargs):
        update_list = [
            "cat_dims",
            "cat_emb_dim",
            "cat_idxs",
            "input_dim",
            "mask_type",
            "n_a",
            "n_d",
            "n_independent",
            "n_shared",
            "n_steps",
        ]
        for var_name, value in kwargs.items():
            if var_name in update_list:
                try:
                    exec(f"global previous_val; previous_val = self.{var_name}")
                    if previous_val != value:  # noqa
                        wrn_msg = f"Pretraining: {var_name} changed from {previous_val} to {value}"  # noqa
                        warnings.warn(wrn_msg)
                        exec(f"self.{var_name} = value")
                except AttributeError:
                    exec(f"self.{var_name} = value")



import os
import pathlib 
from typing import *

import torch
import numpy as np 
import pandas as pd 
import anndata as an
import warnings 
from functools import cached_property

import pytorch_lightning as pl
from pytorch_lightning.loggers import WandbLogger
from pytorch_lightning.callbacks.early_stopping import EarlyStopping
from sklearn.preprocessing import LabelEncoder 

from data import generate_dataloaders, compute_class_weights

import sys, os 
from os.path import join

from networking import upload, download_raw_expression_matrices

here = pathlib.Path(__file__).parent.absolute()

class UploadCallback(pl.callbacks.Callback):
    """Custom PyTorch callback for uploading model checkpoints to the braingeneers S3 bucket.
    
    Parameters:
    path: Local path to folder where model checkpoints are saved
    desc: Description of checkpoint that is appended to checkpoint file name on save
    upload_path: Subpath in braingeneersdev/jlehrer/ to upload model checkpoints to
    """
    
    def __init__(
        self, 
        path: str, 
        desc: str, 
        upload_path='model_checkpoints',
        epochs: int=10,
    ) -> None:
        super().__init__()
        self.path = path 
        self.desc = desc
        self.upload_path = upload_path
        self.epochs = epochs 

    def on_train_epoch_end(self, trainer, pl_module):
        epoch = trainer.current_epoch

        if epoch % self.epochs == 0 and epoch > 0: # Save every ten epochs
            checkpoint = f'checkpoint-{epoch}-desc-{self.desc}.ckpt'
            trainer.save_checkpoint(os.path.join(self.path, checkpoint))
            print(f'Uploading checkpoint at epoch {epoch}')
            
            upload(
                os.path.join(self.path, checkpoint),
                os.path.join('jlehrer', self.upload_path, checkpoint)
            )

class DataModule(pl.LightningDataModule):
    def __init__(
        self, 
        class_label: str,
        datafiles: List[str]=None,
        labelfiles: List[str]=None,
        urls: Dict[str, List[str]]=None,
        sep: str=None,
        unzip: bool=True,
        datapath: str=None,
        assume_numeric_label: bool=True,
        batch_size=32,
        num_workers=0,
        device=('cuda:0' if torch.cuda.is_available() else None),
        *args,
        **kwargs,
    ):
        """
        Creates the DataModule for PyTorch-Lightning training.

        This either takes a dictionary of URLs with the format 
            urls = {dataset_name.extension: 
                        [
                            datafileurl,
                            labelfileurl,
                        ]
                    }

        OR two lists containing the absolute paths to the datafiles and labelfiles, respectively.

        :param class_label: Class label to train on. Must be in all label files 
        :type class_label: str
        :param datafiles: List of absolute paths to datafiles, if not using URLS. defaults to None
        :type datafiles: List[str], optional
        :param labelfiles: List of absolute paths to labelfiles, if not using URLS. defaults to None
        :type labelfiles: List[str], optional
        :param urls: Dictionary of URLS to download, as specified in the above docstring, defaults to None
        :type urls: Dict[str, List[str, str]], optional
        :param unzip: Boolean, whether to unzip the datafiles in the url, defaults to False
        :type unzip: bool, optional
        :param sep: Separator to use in reading in both datafiles and labelfiles. WARNING: Must be homogeneous between all datafile and labelfiles, defaults to '\t'
        :type sep: str, optional
        :param datapath: Path to local directory to download datafiles and labelfiles to, if using URL. defaults to None
        :type datapath: str, optional
        :param assume_numeric_label: If the class_label column in all labelfiles is numeric. Otherwise, we automatically apply sklearn.preprocessing.LabelEncoder to the intersection of all possible labels, defaults to True
        :type assume_numeric_label: bool, optional
        :raises ValueError: If both a dictionary of URL's is passed and labelfiles/datafiles are passed. We can only handle one, not a mix of both, since there isn't a way to determine easily if a string is an external url or not. 

        """    
        super().__init__()

        # Make sure we don't have datafiles/labelfiles AND urls at start
        if urls is not None and datafiles is not None or urls is not None and labelfiles is not None:
            raise ValueError("Either a dictionary of data to download, or paths to datafiles and labelfiles are supported, but not both.")

        self.device = device 
        self.class_label = class_label
        self.urls = urls 
        self.unzip = unzip 
        self.datapath = (
            datapath if datapath is not None else join(here, '..', '..', '..', 'data', 'raw')
        )
        self.assume_numeric_label = assume_numeric_label
        self.batch_size = batch_size
        self.num_workers = num_workers

        # If we have a list of urls, we can generate the list of paths of datafiles/labelfiles that will be downloaded after self.prepare_data()
        if self.urls is not None:
            self.datafiles = [join(self.datapath, f) for f in self.urls.keys()] 
            self.labelfiles = [join(self.datapath, f'labels_{f}') for f in self.urls.keys()] 
        else:
            self.datafiles = datafiles 
            self.labelfiles = labelfiles

        # Warn user in case tsv/csv ,/\t don't match, this can be annoying to diagnose
        suffix = pathlib.Path(self.labelfiles[0]).suffix
        if (sep == '\t' and suffix == 'csv') or (sep == ',' and suffix == '.tsv'):
            warnings.warn(f'Passed delimiter {sep = } doesn\'t match file extension, continuing...')

        # Infer sep based on .csv/.tsv of labelfile (assumed to be homogeneous in case of delimited datafiles) if sep is not passed
        if sep is None:
            if suffix == '.tsv':
                self.sep = '\t'
            elif suffix == '.csv':
                self.sep = ','
            else:
                warnings.warn(f'Separator not passed and not able to be inferred from {suffix=}. Falling back to ","')
                self.sep = ','
        else:
            self.sep = sep 

        self.args = args 
        self.kwargs = kwargs

    def prepare_data(self):
        if self.urls is not None:
            download_raw_expression_matrices(
                self.urls,
                unzip=self.unzip,
                sep=self.sep,
                datapath=self.datapath,
            )
        
        if not self.assume_numeric_label:
            print('assume_numeric_label=False, using sklearn.preprocessing.LabelEncoder and encoding target variables.')

            unique_targets = list(
                set(
                    np.concatenate(
                        [pd.read_csv(df, sep=self.sep).loc[:, self.class_label].unique() for df in self.labelfiles]
                    )
                )
            )

            le = LabelEncoder()
            le = le.fit(unique_targets)
            
            for idx, file in enumerate(self.labelfiles):
                print(f'Transforming labelfile {idx + 1}/{len(self.labelfiles)}')

                labels = pd.read_csv(file, sep=self.sep)
                
                if f'categorical_{self.class_label}' not in labels.columns:
                    labels.loc[:, f'categorical_{self.class_label}'] = labels.loc[:, self.class_label]

                    labels.loc[:, self.class_label] = le.transform(
                        labels.loc[:, f'categorical_{self.class_label}']
                    )

                    labels.to_csv(file, index=False, sep=self.sep) # Don't need to re-index here 

    def setup(self, stage: Optional[str] = None):
        print('Creating train/val/test DataLoaders...')
        trainloader, valloader, testloader = generate_dataloaders(
            datafiles=self.datafiles,
            labelfiles=self.labelfiles,
            class_label=self.class_label,
            sep=self.sep,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=True, # For gpu training
            *self.args,
            **self.kwargs,
        )

        print('Done, continuing to training.')
        self.trainloader = trainloader
        self.valloader = valloader
        self.testloader = testloader
        
        print('Calculating weights')
        self.weights = compute_class_weights(
            labelfiles=self.labelfiles, 
            class_label=self.class_label, 
            sep=self.sep, 
            device=self.device,
        )

    def train_dataloader(self):
        return self.trainloader

    def val_dataloader(self):
        return self.valloader

    def test_dataloader(self):
        return self.testloader

    @cached_property
    def num_labels(self):
        val = []
        for file in self.labelfiles:
            val.extend(list(pd.read_csv(file, sep=self.sep).loc[:, self.class_label].values))

        return len(set(val))

    @cached_property
    def num_features(self):
        if self.urls is not None and not os.path.isfile(self.datafiles[0]):
            print('Trying to calcuate num_features before data has been downloaded. Downloading and continuing...')
            self.prepare_data()

        if 'refgenes' in self.kwargs:
            return len(self.kwargs['refgenes'])
        elif hasattr(self, 'trainloader'):
            return next(iter(self.trainloader))[0].shape[1]
        elif pathlib.Path(self.datafiles[0]).suffix == '.h5ad':
            return an.read_h5ad(self.datafiles[0]).X.shape[1]
        else:
            return pd.read_csv(self.datafiles[0], nrows=1, sep=self.sep).shape[1]

    @cached_property
    def input_dim(self):
        return self.num_features

    @cached_property
    def output_dim(self):
        return self.num_labels

    def __len__(self):
        l = [pd.read_csv(f, sep=self.sep).shape[0] for f in self.labelfiles]

        return sum(l)

def generate_trainer(
    datafiles: List[str],
    labelfiles: List[str],
    class_label: str,
    batch_size: int,
    num_workers: int,
    optim_params: Dict[str, Any]={
        'optimizer': torch.optim.Adam,
        'lr': 0.02,
    },
    weighted_metrics: bool=None,
    scheduler_params: Dict[str, float]=None,
    wandb_name: str=None,
    weights: torch.Tensor=None,
    max_epochs=500,
    *args,
    **kwargs,
):
    """
    Generates PyTorch Lightning trainer and datasets for model training.

    :param datafiles: List of absolute paths to datafiles
    :type datafiles: List[str]
    :param labelfiles: List of absolute paths to labelfiles
    :type labelfiles: List[str]
    :param class_label: Class label to train on 
    :type class_label: str
    :param weighted_metrics: To use weighted metrics in model training 
    :type weighted_metrics: bool
    :param batch_size: Batch size in dataloader
    :type batch_size: int
    :param num_workers: Number of workers in dataloader
    :type num_workers: int
    :param optim_params: Dictionary defining optimizer and any needed/optional arguments for optimizer initializatiom
    :type optim_params: Dict[str, Any]
    :param wandb_name: Name of run in Wandb.ai, defaults to ''
    :type wandb_name: str, optional
    :return: Trainer, model, datamodule 
    :rtype: Trainer, model, datamodule 
    """

    device = ('cuda:0' if torch.cuda.is_available() else 'cpu')
    print(f'Device is {device}')
    
    here = pathlib.Path(__file__).parent.absolute()
    data_path = os.path.join(here, '..', '..', '..', 'data')

    wandb_logger = WandbLogger(
        project=f"tabnet-classifer-sweep",
        name=wandb_name
    )

    uploadcallback = UploadCallback(
        path=os.path.join(here, 'checkpoints'),
        desc=wandb_name
    )

    early_stop_callback = EarlyStopping(
        monitor=("weighted_val_accuracy" if weighted_metrics else "val_accuarcy"), 
        min_delta=0.00, 
        patience=3, 
        verbose=False, 
        mode="max"
    )

    module = DataModule(
        datafiles=datafiles, 
        labelfiles=labelfiles, 
        class_label=class_label, 
        batch_size=batch_size,
        num_workers=num_workers,
    )

    model = SIMSClassifier(
        input_dim=module.num_features,
        output_dim=module.num_labels,
        weighted_metrics=weighted_metrics,
        optim_params=optim_params,
        scheduler_params=scheduler_params,
        weights=weights,
    )
    
    trainer = pl.Trainer(
        gpus=(1 if torch.cuda.is_available() else 0),
        auto_lr_find=False,
        # gradient_clip_val=0.5,
        logger=wandb_logger,
        max_epochs=max_epochs,
        callbacks=[
            uploadcallback, 
        ],
        val_check_interval=0.25, # Calculate validation every quarter epoch instead of full since dataset is large, and would like to test this 
    )

    return trainer, model, module




from typing import *

import torch 
import numpy as np 
from torchmetrics.functional import *
import torchmetrics 
from tqdm import tqdm
import sys

sys.path.append('../src')
from data import *
from lightning_train import *
from model import *
from functools import partial 

def confusion_matrix(model, dataloader, num_classes):
    confusion_matrix = torch.zeros(num_classes, num_classes)
    with torch.no_grad():
        for i, (inputs, classes) in enumerate(tqdm(dataloader)):
            outputs, _ = model(inputs)
            
            _, preds = torch.max(outputs, 1)
            for t, p in zip(classes.view(-1), preds.view(-1)):
                    confusion_matrix[t.long(), p.long()] += 1
                    
    return confusion_matrix 

def median_f1(tps, fps, fns):
    precisions = tps / (tps+fps)
    recalls = tps / (tps+fns)
    
    f1s = 2*(np.dot(precisions, recalls)) / (precisions + recalls)
    
    return np.nanmedian(f1s)

def per_class_f1(*args, **kwargs):
    res = torchmetrics.functional.f1_score(*args, **kwargs, average='none')
    return res

def per_class_precision(*args, **kwargs):
    res = torchmetrics.functional.precision(*args, **kwargs, average='none')
    
    return res

def per_class_recall(*args, **kwargs):
    res = torchmetrics.functional.precision(*args, **kwargs, average='none')
    
    return res 

def weighted_accuracy(*args, **kwargs):
    res = torchmetrics.functional.accuracy(*args, **kwargs, average='weighted')
    
    return res 

def balanced_accuracy(*args, **kwargs):
    res = torchmetrics.functional.accuracy(*args, **kwargs, average='macro')
    
    return res 

def aggregate_metrics(num_classes) -> Dict[str, Callable]:
    metrics = {
        # Accuracies
        'total_accuracy': torchmetrics.functional.accuracy,
        'balanced_accuracy': partial(balanced_accuracy, num_classes=num_classes),
        'weighted_accuracy': partial(weighted_accuracy, num_classes=num_classes),
        
        # Precision, recall and f1s
        'precision': torchmetrics.functional.precision,
        'recall': torchmetrics.functional.recall,
        'f1': torchmetrics.functional.f1_score,
        
        # Per class 
        'per_class_f1': partial(per_class_f1, num_classes=num_classes),
        'per_class_precision': partial(per_class_precision, num_classes=num_classes),
        'per_class_recall': partial(per_class_recall, num_classes=num_classes),
        
        # Random stuff I might want
        'specificity': partial(torchmetrics.functional.specificity, num_classes=num_classes),
        'confusion_matrix': partial(torchmetrics.functional.confusion_matrix, num_classes=num_classes),
        'auroc': partial(torchmetrics.functional.auroc, num_classes=num_classes)
    }
    
    return metrics 


import pathlib 
import os 
import boto3 
import urllib 
from typing import * 

here = pathlib.Path(__file__).parent.absolute()
data_path = os.path.join(here, '..', '..', 'data')

with open(os.path.join(here, '..', 'credentials')) as f:
    key, access = [line.rstrip() for line in f.readlines()]

s3 = boto3.resource(
    's3',
    endpoint_url="https://s3-west.nrp-nautilus.io/",
    aws_access_key_id=key,
    aws_secret_access_key=access,
)

def _download_from_key(
    key: str, 
    localpath: str,
) -> None:
    """Helper function that downloads all files recursively from the given key (folder) from the braingeneersdev S3 bucket

    :param key: S3 folder (key) to start downloading recursively from
    :type key: str
    :param localpath: Optional argument, downloads to a subfolder under the data/processed/ folder # TODO add folder generation
    :type localpath: str 
    """    

    print(f'Key is {key}')
    reduced_files = list_objects(key)

    if not os.path.exists(localpath):
        print(f'Download path {localpath} doesn\'t exist, creating...')
        os.makedirs(localpath, exist_ok=True)

    for f in reduced_files:
        if not os.path.isfile(os.path.join(data_path, 'processed', f.split('/')[-1])):
            print(f'Downloading {f} from S3')
            download(
                f,
                os.path.join(localpath, f.split('/')[-1]) # Just the file name in the list of objects
            )

def download_clean_from_s3(
    file: str=None,
    local_path: str=None,
) -> None:
    """Downloads the cleaned data from s3 to be used in model training

    :param file: file name to download from braingeneersdev S3 bucket, defaults to None
    :type file: str, optional
    :param local_path: path to download file to, defaults to None
    :type local_path: str, optional
    """    
    os.makedirs(os.path.join(data_path, 'processed'), exist_ok=True)
    if not file: # No single file passed, so download recursively
        print('Downloading all clean data...')
        key = os.path.join('jlehrer', 'expression_data', 'processed')
        local_path = os.path.join(data_path, 'processed')

        _download_from_key(key, local_path) 
    else:
        print(f'Downloading {file} from clean data')
        local_path = (os.path.join(data_path, 'processed', file) if not local_path else local_path)
        download(
            os.path.join('jlehrer', 'expression_data', 'processed', file),
            local_path
        )

def download_interim_from_s3(
    file: str=None,
    local_path: str=None,
) -> None:
    """Downloads the interim data from S3. Interim data is in the correct structural format but has not been cleaned

    :param file: _description_, defaults to None
    :type file: str, optional
    :param local_path: _description_, defaults to None
    :type local_path: str, optional
    """
    os.makedirs(os.path.join(data_path, 'interim'), exist_ok=True)

    if not file:
        print('Downloading all interim data')
        key = os.path.join('jlehrer', 'expression_data', 'interim')
        local_path = os.path.join(data_path, 'interim')
        _download_from_key(key, local_path)
    else:
        print(f'Downloading {file} from interim data')
        local_path = (os.path.join(data_path, 'interim', file) if not local_path else local_path)
        download(
            os.path.join('jlehrer', 'expression_data', 'interim', file),
            local_path,
        )
        
def download_raw_from_s3(
    file: str=None,
    local_path: str=None,
) -> None:
    """Downloads the raw expression matrices from s3


    :param file: _description_, defaults to None
    :type file: str, optional
    :param local_path: _description_, defaults to None
    :type local_path: str, optional
    """
    os.makedirs(os.path.join(data_path, 'raw'), exist_ok=True)
    if not file: 
        print('Downloading all raw data')
        key = os.path.join('jlehrer', 'expression_data', 'raw')
        local_path = os.path.join(data_path, 'raw')
        _download_from_key(key, local_path)
    else:
        print(f'Downloading {file} from raw data')
        local_path = (os.path.join(data_path, 'raw', file) if not local_path else local_path)
        download(
            os.path.join('jlehrer', 'expression_data', 'raw', file), 
            local_path
        )

def download(remote_name, file_name=None) -> None:
    """
    Downloads a file from the braingeneersdev S3 bucket 

    Parameters:
    remote_name: S3 key to download. Must be a single file
    file_name: File name to download to. Default is remote_name
    """
    if file_name == None:
        file_name == remote_name

    s3.Bucket('braingeneersdev').download_file(
        Key=remote_name,
        Filename=file_name,
    )

def download_raw_expression_matrices(
    datasets: Dict[str, Tuple[str, str]],
    upload: bool=False,
    unzip: bool=True,
    datapath: str=None
) -> None:
    """Downloads all raw datasets and label sets from cells.ucsc.edu, and then unzips them locally

    :param datasets: Dictionary of datasets such that each key maps to a tuple containing the expression matrix csv url in the first element,
                    and the label csv url in the second url, defaults to None
    :type datasets: Dict[str, Tuple[str, str]], optional
    :param upload: Whether or not to also upload data to the braingeneersdev S3 bucket , defaults to False
    :type upload: bool, optional
    :param unzip: Whether to also unzip expression matrix, defaults to False
    :type unzip: bool, optional
    :param datapath: Path to folder to download data to. Otherwise, defaults to data/
    :type datapath: str, optional
    """    
    # {local file name: [dataset url, labelset url]}
    data_path = (datapath if datapath is not None else os.path.join(here, '..', '..', '..', 'data'))

    for file, links in datasets.items():
        datafile_path = os.path.join(data_path, 'raw', file)

        labelfile = f'{file[:-4]}_labels.tsv'

        datalink, _ = links

        # First, make the required folders if they do not exist 
        for dir in 'raw':
            os.makedirs(os.path.join(data_path, dir), exist_ok=True)

        # Download and unzip data file if it doesn't exist 
        if not os.path.isfile(datafile_path):
            print(f'Downloading zipped data for {file}')
            urllib.request.urlretrieve(
                datalink,
                f'{datafile_path}.gz',
            )

            if unzip:
                print(f'Unzipping {file}')
                os.system(
                    f'zcat < {datafile_path}.gz > {datafile_path}'
                )

                print(f'Deleting compressed data')
                os.system(
                    f'rm -rf {datafile_path}.gz'
                )


        # If upload boolean is passed, also upload these files to the braingeneersdev s3 bucket
        if upload:
            print(f'Uploading {file} and {labelfile} to braingeneersdev S3 bucket')
            upload(
                datafile_path,
                os.path.join('jlehrer', 'expression_data', 'raw', file)
            )
def upload(file_name, remote_name=None) -> None:
    """
    Uploads a file to the braingeneersdev S3 bucket
    
    Parameters:
    file_name: Local file to upload
    remote_name: Key for S3 bucket. Default is file_name
    """
    if remote_name == None:
        remote_name = file_name

    s3.Bucket('braingeneersdev').upload_file(
        Filename=file_name,
        Key=remote_name,
)

def list_objects(prefix: str) -> list:
    """
    Lists all the S3 objects from the braingeneers bucket with the given prefix.

    Parameters:
    prefix: Prefix to filter S3 objects. For example, if we want to list all the objects under 'jlehrer/data' we pass 
            prefix='jlehrer/data'

    Returns:
    List[str]: List of objects with given prefix
    """

    objs = s3.Bucket('braingeneersdev').objects.filter(Prefix=prefix)
    return [x.key for x in objs]
